! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module adsmooth_mod 


  use numkind  
  use fitsimage
  use detect_io


  implicit none


  contains

  !----------------------------------------

  

  subroutine fkgauss(sigma,ixsize,iysize,ffkernel,norm,ierr)
  
   !---------------------------------------------
   ! calculate Fourier transformed of gaussian kernel
   ! rearrange quadrants for FFTW input:
   !    1|2           4|3
   !    ---    =>     ---
   !    3|4           2|1 
   !  currently only working for ixsize=iysize
   !---------------------------------------------


    real(kind=real32)                          :: sigma    
    complex(kind=4), dimension(:,:),pointer    :: ffkernel
    real(kind=real32),dimension(:,:), allocatable :: kernel
    integer(kind=int32)                        :: ixsize 
    integer(kind=int32)                        :: iysize 
    integer(kind=int32)                        :: ierr

    integer(kind=int32)                        :: i,j,ii,jj,ii2,jj2,hsize,ncalc
    real(kind=real32)                          :: x,y,twosig2,norm,sig_nux
 
    type(r4image_type)  :: fkima
    integer(kind=int32)                        :: status
      
    real(kind=real32)                          :: pi=3.141592654,twopi=6.283185308   
    character(len=*), parameter                :: rtname='FKGAUSS'
   
    ierr=0


    hsize=ixsize/2
  
!    allocate(ffkernel(hsize+1,iysize)) 
    allocate(kernel(hsize+1,iysize)) 
    kernel(:,:)=0.0
    sig_nux = real(ixsize)/(2.*pi*sigma)

    ncalc=min(int(5.*sig_nux),hsize)
    twosig2 = 2.0*sig_nux*sig_nux
    norm=1./(pi * 2.0*sig_nux*sig_nux)
    do j=1,ncalc
      y=real(j)-1.0
      jj=iysize-j+2
      do  i=1,ncalc
        ii=ixsize-i+1
        x=real(i)-1.0
        kernel(i,j)   = exp(-(x*x+y*y)/twosig2)
        if (j > 1) kernel(i,jj)  = kernel(i,j)
      enddo
    enddo
    



      ffkernel=cmplx(kernel,kernel)
    
      allocate(fkima%arr(hsize+1,iysize))
      fkima%arr=real(kernel)

      fkima%shapehdr%naxis1=hsize+1
      fkima%shapehdr%naxis2=iysize
      call fput_r4image('kernel0.fits',fkima,status) 



  end  subroutine fkgauss

  !--------------------------------------------------------------------------------------------




  subroutine kgauss(sigma,ixsize,iysize,kernel,norm,ierr)
  
   !---------------------------------------------
   ! calculate gaussian kernel
   ! rearrange quadrants for FFTW input:
   !    1|2           4|3
   !    ---    =>     ---
   !    3|4           2|1 
   !---------------------------------------------


    real(kind=real32)                        :: sigma    
    real(kind=real32), dimension(:,:)        :: kernel
    integer(kind=int32)                      :: ixsize 
    integer(kind=int32)                      :: iysize 
    integer(kind=int32)                      :: ierr

    integer(kind=int32)                      :: i,j,ii,jj,ii2,jj2,ij,ji,hsize,ncalc
    real(kind=real32)                        :: x,y,twosig2,norm
       
    real(kind=real32)                        :: pi=3.141592654,twopi=6.283185308   
    character(len=*), parameter              :: rtname='KGAUSS'
   
    ierr=0




    hsize=min(iysize/2,ixsize/2)
    ncalc=min(int(5.*sigma),hsize)
    twosig2 = 2.0*sigma*sigma
    norm=1./(pi * 2.0*sigma*sigma)
    do j=1,ncalc
      y=real(j)-0.5
!      jj=j+hsize
!      jj2=hsize-j+1
      ! sorting for FFT input:
      jj=j
      jj2=iysize-j+1
      ij=ixsize-j+1
      do  i=1,j
        x=real(i)-0.5
!        ii=i+hsize
!        ii2=hsize-i+1
        ! sorting for FFT input:
        ii=i
        ii2=ixsize-i+1
        ji=iysize-i+1
        kernel(ii,jj)   = norm*exp(-(x*x+y*y)/twosig2) ! Octant 1
        kernel(jj,ii)   = kernel(ii,jj)  ! Oct 2
        kernel(ii2,jj)  = kernel(ii,jj)  ! Oct 8
        kernel(ij,ii)   = kernel(ii,jj)  ! Oct 7
        kernel(ii2,jj2) = kernel(ii,jj)  ! Oct 6 
        kernel(ij,ji)   = kernel(ii,jj)  ! Oct 5
        kernel(ii,jj2)  = kernel(ii,jj)  ! Oct 3
        kernel(jj,ji)   = kernel(ii,jj)  ! Oct 4      
      enddo
    enddo

    
  end  subroutine kgauss

  !--------------------------------------------------------------------------------------------

  subroutine fconvolve(sigma,image1,image2,kernel,smoothed1,smoothed2,ixsize,iysize,ierr)

    real(kind=real32)                   :: sigma
    real(kind=real32), dimension(:,:)   :: image1
    real(kind=real32), dimension(:,:)   :: image2               
    real(kind=real32), dimension(:,:)   :: smoothed1   
    real(kind=real32), dimension(:,:)   :: smoothed2  
    real(kind=real32), dimension(:,:)   :: kernel


    integer(kind=int32)                 :: ixsize,hxsize
    integer(kind=int32)                 :: iysize,hysize
    integer(kind=int32)                 :: ierr

    integer(kind=int32)                 :: status


    real(kind=real32)                    :: pi=3.141592654,twopi=6.283185308   
    real(kind=real32)                    :: norm


    type(r4image_type)  :: fkima

    complex(kind=4), dimension(:,:), allocatable       :: ffin, ffout
    complex(kind=4), dimension(:,:), allocatable       :: ffkernel,ffmul,ffexpmul
    complex(kind=4), dimension(:,:), allocatable, save :: ffima,ffexp
    real(kind=real32), dimension(:,:), allocatable     :: re_in,re_out


    integer(kind=int64),save       :: implan, explan, kplan, fplan,fexplan,rplan
    real(kind=real32)              :: fftscale
!    real(kind=real32)              :: t1,t2,tcpu
    integer(kind=int32), parameter :: fftw_measure=0
    integer(kind=int32), parameter :: fftw_exhaustive=8
    integer(kind=int32), parameter :: fftw_patient=32
    integer(kind=int32), parameter :: fftw_estimate=64
    integer(kind=int32), parameter :: fftw_estimate_patient=128

    integer(kind=int32)            :: plan_type
 
    logical,save                    :: first=.true.

    character(len=*), parameter    :: rtname='fconvolve'



    fftscale=1./float((ixsize*iysize))

    ierr=0

    status=0


     hxsize=ixsize/2
     hysize=iysize/2
 

    if (.not. allocated(re_in))  allocate(re_in(ixsize,iysize),stat=status)
    if (.not. allocated(re_out)) allocate(re_out(ixsize,iysize),stat=status)
    if (.not. allocated(ffin))   allocate(ffin(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffout))  allocate(ffout(1:hxsize+1,1:iysize),stat=status)
   


    if (.not. allocated(ffkernel)) allocate(ffkernel(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffmul))    allocate(ffmul(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffexpmul)) allocate(ffexpmul(1:hxsize+1,1:iysize),stat=status)

    if (first) then 
      allocate(ffima(1:hxsize+1,1:iysize),stat=status) 
      allocate(ffexp(1:hxsize+1,1:iysize),stat=status)
    endif



    if (status == 0) then
    
!      call cpu_time(t1)

      ! note: FFTW_MEASURE modifies input array
      plan_type=FFTW_ESTIMATE
      !plan_type=FFTW_MEASURE


      call sfftw_plan_dft_r2c_2d(rplan,ixsize,iysize,re_in,ffout,plan_type)
      call sfftw_plan_dft_c2r_2d(fplan,ixsize,iysize,ffin,re_out,plan_type)  


!      call cpu_time(t2)
!      write(*,*) 'CPU time (plans): ',t2-t1
  
  !   call fkgauss(sigma,ixsize,iysize,ffkernel,norm,ierr)


      if (first) then 
        re_in=image1
        call sfftw_execute(rplan)
        ffima=ffout

        re_in=image2
        call sfftw_execute(rplan)
        ffexp=ffout
      endif

      re_in=kernel
      call sfftw_execute(rplan)
      ffkernel=ffout
     


      ffmul    = ffima*ffkernel*fftscale
      ffexpmul = ffexp*ffkernel*fftscale

   !  !$OMP PARALLEL
   !  !$OMP SECTIONS PRIVATE(ffin,re_out)
   !  !$OMP SECTION
      ffin = ffmul
      call sfftw_execute(fplan)
      smoothed1 = re_out
    !  !$OMP SECTION     
      ffin = ffexpmul
      call sfftw_execute(fplan)
      smoothed2 = re_out
   !  !$OMP END SECTIONS
   !  !$OMP END PARALLEL

!      call  cpu_time(t1)
!      write(*,*) 'CPU time (fold): ',t1-t2

      first=.false.

    else
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif 
   
    
  
  end subroutine fconvolve

end module adsmooth_mod 
